/***************************************************************************
 * Artekit STM32 DIY lightsaber audio board
 * Copyright (C) 2013 Artekit Italy
 * Written by Ivan Meleca
 * http://www.artekit.eu
 
### dma.c
 
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************/

#include "dma.h"
#include "sys.h"

/* Max quantity of requests */
#define DMA_MAX_REQUESTS	16

/* Dummy for DMA */
static uint8_t dummy = 0;

/* This is set to 1 if the DMA is working on a request */
static volatile uint8_t on_req = 0;

/* This array keeps track of the DMA requests */
DMA_REQUEST dma_requests[DMA_MAX_REQUESTS];

/* Pointer to the currently running request */
DMA_REQUEST* dma_running = 0;

/* Gets a free request from the dma_requests array */
DMA_REQUEST* GetFreeReq(void)
{
	uint8_t i;
	
	for (i = 0; i < DMA_MAX_REQUESTS; i++)
	{
		if (!dma_requests[i].active)
		{
			dma_requests[i].active = 1;
			return &dma_requests[i];
		}
	}
	
	return 0;
}

/* Appends a request */
static void AppendReq(DMA_REQUEST* req)
{
	DMA_REQUEST* next;
	DMA_REQUEST* last = 0;

	__disable_irq();
	next = dma_running;
	__enable_irq();
	
	if (!next) return;
	
	while (next)
	{
		last = next;
		next = next->next;
	}
	
	last->next = req;
	req->next = 0;
}

/* Starts a DMA reading from the serial flash */
static void ReadDMA(uint8_t* buffer, uint32_t len)
{
	/* Wait for the bus to be free */
	while (!(SPI1->SR & SPI_SR_TXE));
	while (SPI1->SR & SPI_SR_BSY);
	
	/* SPI transmission DMA (from a dummy address to SPI1->DR) */
    DMA1_Channel3->CPAR = (uint32_t) &(SPI1->DR);
    DMA1_Channel3->CMAR = (uint32_t) &dummy;
    DMA1_Channel3->CNDTR = len;
    DMA1_Channel3->CCR = DMA_CCR1_DIR;

    /* SPI reception DMA (from SPI1->DR to our buffer) */
    DMA1->IFCR = DMA_IFCR_CGIF2;
    DMA1->IFCR = 0;
    DMA1_Channel2->CPAR = (uint32_t) &(SPI1->DR);
    DMA1_Channel2->CMAR = (uint32_t) buffer;
    DMA1_Channel2->CNDTR = len;
    DMA1_Channel2->CCR = (1 << 12) | DMA_CCR1_MINC | DMA_CCR1_TCIE;
	
	NVIC_EnableIRQ(DMA1_Channel2_IRQn);
	NVIC_SetPriority(DMA1_Channel2_IRQn, 14);

	/* Clear RX flags */
    SPI_I2S_ReceiveData(SPI1);
	SPI_I2S_GetFlagStatus(SPI1, 0);
	
	/* Enable DMA */
    DMA1_Channel2->CCR |= DMA_CCR1_EN;
    DMA1_Channel3->CCR |= DMA_CCR1_EN;
    SPI1->CR2 |= SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;   
}

void DMA1_Channel3_IRQHandler(void)
{
	if (DMA_GetITStatus(DMA1_IT_TC3))
	{
		DMA_ClearITPendingBit(DMA1_IT_TC3);
		
		while (!(SPI1->SR & SPI_SR_TXE));
		while (SPI1->SR & SPI_SR_BSY);
		
		SPI1->CR2 &= ~SPI_CR2_TXDMAEN;
		DMA1_Channel3->CCR &= ~DMA_CCR1_EN;
		NVIC_DisableIRQ(DMA1_Channel3_IRQn);
	    ReadDMA(dma_running->buffer, dma_running->qty);
	}
}

void LaunchDMA(DMA_REQUEST* req)
{
	on_req = 1;
	
	dma_running = req;
		
	while (!(SPI1->SR & SPI_SR_TXE));
	while (SPI1->SR & SPI_SR_BSY);
	
	FLASH_CS_ON();
	
	SPI_I2S_ReceiveData(SPI1);
	SPI_I2S_GetFlagStatus(SPI1, 0);
	
	SPI1->CR1 &= ~0x38;
	
	DMA1->IFCR = DMA_IFCR_CGIF2;
    DMA1->IFCR = 0;
	
	DMA_ClearITPendingBit(DMA1_IT_TC3 | DMA1_IT_HT3 | DMA1_IT_GL3 | DMA1_IT_TE3);
	
	DMA1_Channel3->CPAR = (uint32_t) &(SPI1->DR);
    DMA1_Channel3->CMAR = (uint32_t) req->cmd;
	
	DMA1_Channel3->CNDTR = 5;

    DMA1_Channel3->CCR = DMA_CCR1_DIR | DMA_CCR1_MINC | DMA_CCR1_TCIE;
	
	NVIC_EnableIRQ(DMA1_Channel3_IRQn);
	
	DMA1_Channel3->CCR |= DMA_CCR1_EN;
	SPI1->CR2 |= SPI_CR2_TXDMAEN;
}

void DMA1_Channel2_IRQHandler(void)
{
	if (DMA_GetITStatus(DMA1_IT_TC2))
	{
		DMA_ClearITPendingBit(DMA1_IT_TC2);
		SPI1->CR2 &= ~(SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN);
		DMA1_Channel3->CCR &= ~DMA_CCR1_EN;
		DMA1_Channel2->CCR &= ~DMA_CCR1_EN;
		
		FLASH_CS_OFF();
		
		dma_running->active = 0;
		*dma_running->signal_ptr = 1;
		if (dma_running->next && dma_running->next->active)
		{
			LaunchDMA(dma_running->next);
		} else {
			dma_running = 0; 
			on_req = 0;
		}
	}
}

void ProcessRequest(DMA_REQUEST* req)
{
	if (!on_req)
	{
		LaunchDMA(req);
	} else {
		AppendReq(req);
	}
}
